;; mercury.el --- major mode for editing and running Mercury code
;;========================================================================
;;  Frankly this is no more than the well-known Prolog ruthlessly
;;  adapted for Mercury usage, chiefly better syntax highlighting.
;;  Accordingly,
;;  o  credits remain at the original authors of Prolog Mode (see below);
;;  o  for Prolog code, this mode will be probaly useless 
;;
;;  Nick Rudnick <nick.rudnick@googlemail.com>

;;========================================================================
;; Copyright (C) 1986, 1987, 1997, 1998, 1999, 2002, 2003 Free Software Foundation, Inc.

;; Authors: Emil strm <emil_astrom@hotmail.com>
;;          Milan Zamazal <pdm@freesoft.cz>
;;          Stefan Bruda <bruda@ubishops.ca>
;;          (see below for more details)
;; Keywords: mercury major mode

;;;;;;;;;;;;;;;;;;;;;;


;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to the
;; Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;; Original author: Masanobu UMEDA <umerin@mse.kyutech.ac.jp>
;; Parts of this file was taken from a modified version of the original
;; by Johan Andersson, Peter Olin, Mats Carlsson, Johan Bevemyr, Stefan
;; Andersson, and Per Danielsson (all SICS people), and Henrik Bkman
;; at Uppsala University, Sweden.
;;
;; Some ideas and also a few lines of code have been borrowed (not stolen ;-)
;; from Oz.el, the Emacs major mode for the Oz programming language,
;; Copyright (C) 1993 DFKI GmbH, Germany, with permission.
;; Authors: Ralf Scheidhauer and Michael Mehl ([scheidhr|mehl]@dfki.uni-sb.de)
;;
;; More ideas and code have been taken from the SICStus debugger mode
;; (http://www.csd.uu.se/~perm/source_debug/index.shtml -- broken link
;; as of Mon May 5 08:23:48 EDT 2003) by Per Mildner.

;;; Commentary:
;;
;; This package provides a major mode for editing Mercury code, with
;; all the bells and whistles one would expect, including syntax
;; highlighting and auto indentation.
;;
;; The code requires the comint, easymenu, info, imenu, and font-lock
;; libraries. These are normally distributed with GNU Emacs and
;; XEmacs.

;;; Installation:
;;
;; Insert the following lines in your init file--typically ~/.emacs
;; (FSF Emacs and XEmacs <21.4), or ~/.xemacs/init.el (XEmacs
;; 21.4)--to use this mode when editing Mercury files under Emacs:
;;
;; (setq load-path (cons "/usr/lib/xemacs/site-lisp" load-path))
;; (autoload 'mercury-mode "mercury" "Major mode for editing Mercury programs." t)
;; (setq auto-mode-alist (append '(("\\.m$" . mercury-mode)) auto-mode-alist))
;;
;; where the path in the first line is the file system path to this file.
;; MSDOS paths can be written like "d:/programs/emacs-19.34/site-lisp". 
;; Note: In XEmacs, either `/usr/lib/xemacs/site-lisp' (RPM default in
;; Red Hat-based distributions) or `/usr/local/lib/xemacs/site-lisp'
;; (default when compiling from sources) are automatically added to
;; `load-path', so the first line is not necessary provided that you
;; put this file in the appropriate place.
;;
;; The last s-expression above makes sure that files ending with .m
;; are assumed to be Mercury files and not Objective C (, which is the default
;; Emacs setting??). If this is not wanted, remove this line. It is then
;; necessary to either
;;
;;  o  insert in your Mercury files the following comment as the first line:
;;
;;       % -*- Mode: Mercury -*-
;;
;;     and then the file will be open in Mercury mode no matter its 
;;     extension, or
;;
;;  o  manually switch to Mercury mode after opening a Mercury file, by typing 
;;     M-x mercury-mode.
;;
;; Note: I (Stefan, the current maintainer) work under XEmacs.  Future
;;   developments will thus biased towards XEmacs (OK, I admit it, I
;;   am biased towards XEmacs in general), though I will do my best
;;   to keep the FSF Emacs compatibility.  So if you work under Emacs
;;   and see something that does not work do drop me a line, as I
;;   have a smaller chance to notice this kind of bugs otherwise.

;; Changelog:

;; Version 1.9
;;  o  Parenthesis expressions are now indented by default so that
;;     components go one underneath the other, just as for compound
;;     terms.  You can use the old style (the second and subsequent
;;     lines being indented to the right in a parenthesis expression)
;;     by setting the customizable variable `mercury-paren-indent-p'
;;     (group "Mercury Indentation") to t.
;;  o  (Somehow awkward) handling of the 0' character escape
;;     sequence. I am looking into a better way of doing it but
;;     prospects look bleak.  If this breaks things for you please let
;;     me know and also set the `mercury-char-quote-workaround' (group
;;     "Mercury Other") to nil.
;; Version 1.8
;;  o  Key binding fix.
;; Version 1.7
;;  o  Fixed a number of issues with the syntax of single quotes,
;;     including Debian bug #324520.
;; Version 1.6
;;  o  Fixed mercury mode menu initialization (Debian bug #226121).
;;  o  Fixed (i.e., eliminated) Delete remapping (Debian bug #229636).
;;  o  Corrected indentation for clauses defining quoted atoms.
;; Version 1.5:
;;  o  Keywords fontifying should work in console mode so this is
;;     enabled everywhere.
;; Version 1.3:
;;  o  Info-follow-nearest-node now called correctly under Emacs too
;;     (thanks to Nicolas Pelletier).  Should be implemented more
;;     elegant (i.e., without compilation warnings) in the future.
;; Version 1.1:
;;  o  Fixed dots in the end of line comments causing indentation
;;     problems.  The following code is now correctly indented (note
;;     the dot terminating the comment):
;;        a(X) :- b(X),                   
;;            c(X).                  % comment here.
;;        a(X).
;;     and so is this (and variants):
;;        a(X) :- b(X),                   
;;            c(X).                  /* comment here. */
;;        a(X).
;; Version 1.0:
;;  o  Revamped the menu system.
;;  o  This is more of a renumbering than a new edition.  I promoted
;;     the mode to version 1.0 to emphasize the fact that it is now
;;     mature and stable enough to be considered production (in my
;;     opinion anyway).
;; Version 0.1.41:
;;  o  FSF Emacs compatibility fixes.
;; Version 0.1.40:
;;  o  mercury-get-predspec is now suitable to be called as
;;     imenu-extract-index-name-function.  The predicate index works.
;;  o  Since imenu works now as advertised, mercury-imenu-flag is t
;;     by default.
;;  o  Eliminated mercury-create-predicate-index since the imenu
;;     utilities now work well.  Actually, this function is also
;;     buggy, and I see no reason to fix it since we do not need it
;;     anyway.
;;  o  Fixed mercury-pred-start, mercury-clause-start, mercury-clause-info.
;;  o  Fix for mercury-build-case-strings; now mercury-upper-case-string
;;     and mercury-lower-case-string are correctly initialized, 
;;  o  Various font-lock changes; most importantly, block comments (/*
;;     ... */) are now correctly fontified in XEmacs even when they
;;     extend on multiple lines.
;; Version 0.1.36: 
;;  o  The debug prompt of SWI Prolog is now correctly recognized.
;; Version 0.1.35: 
;;  o  Minor font-lock bug fixes.
;;========================================================================
;; TO-DO:
;; (o)  Something close to a rename refactoring...
;; (o)  Abschluss '"' nach einer '0' misslingt; wieder hinkriegen...
;; erledigt -- s. Z. 563...
(defvar mercury-mode-version "0.0" "Mercury mode version number")
(eval-when-compile
    (require 'compile)
    (require 'font-lock)
    ;; We need imenu everywhere because of the predicate index!
    (require 'imenu)
    (require 'info)
    (require 'shell)
)
(require 'comint)
(require 'easymenu)
(require 'font-lock)
;=====================================================================================

;;; Code:
(defgroup mercury nil
    "Major modes for editing and running Mercury files."
    :group 'languages)
(defgroup mercury-faces nil
    "Mercury mode specific faces."
    :group 'font-lock)
(defgroup mercury-indentation nil
    "Mercury mode indentation configuration."
    :group 'mercury)
(defgroup mercury-font-lock nil
    "Mercury mode font locking patterns."
    :group 'mercury)
(defgroup mercury-keyboard nil
    "Mercury mode keyboard flags."
    :group 'mercury)
(defgroup mercury-inferior nil
    "Inferior Mercury mode options."
    :group 'mercury)
(defgroup mercury-other nil
    "Other Mercury mode options."
    :group 'mercury)




;; Set everything up
(defun mercury-font-lock-keywords ()
    "Set up font lock keywords for Mercury."
    ;(when window-system
    
    ;; Font Lock Patterns
    (let ( ;=====================================================================
        ;; "Native" Mercury patterns
        (head-predicates
            '("^\\([a-z][a-zA-Z0-9_.]*\\)" 0 'font-lock-function-name-face))
		(variables
            '("\\<\\([_A-Z][a-zA-Z0-9_]*\\)" 1 font-lock-variable-name-face))
		(important-elements
		    (list "[][}{;,|]\\|+\\|(\\|)\\|\\.\\|!\\|@\\|:\\|\\^\\|~\\|&\\|?-\\|<\\|>\\|\\\\\\|/\\|*\\|-\\|=" 0 'font-lock-builtin-face))
		(important-elements-1 '("[^-*]\\(->\\)" 1 font-lock-keyword-face))
        ;;;; TO-DO: Erste Klammern ist nicht gelb, sondern rosa-rot
        ;;;; TO-DO: In Kommentaren auch davor stehende, durch Leerraum getrennte Wrter
        ;;;;        rosarot (z.B. 'as transform_value/4')
		(predspecs                        ; module:predicate/cardinality
		    (list 
			    (format "\\<\\(%s.\\|\\)%s/[0-9]+" mercury-atom-regexp mercury-atom-regexp)
				;"\\<\\([_a-z][_a-zA-Z0-9].\\|\\)[_a-z][_a-zA-Z0-9]/[0-9]+"
                0 font-lock-function-name-face 'prepend))
		(keywords                        ; directives (queries)
		    (list
                (concat
				    "\\<\\("
					(mercury-make-keywords-regexp mercury-keywords-i)
					"\\|"
					(mercury-make-keywords-regexp mercury-determinism-specificators-i)
					"\\)\\>")
				1 'font-lock-keyword-face))
		(quoted_atom 
            (list mercury-quoted-atom-regexp 2 'font-lock-string-face 'append))
		(string_operator
		    (list mercury-string-operator-regexp 1 'font-lock-builtin-face 'append))
		(string 
            (list  mercury-string-regexp 1 'font-lock-string-face 'append))
		;; Mercury specific patterns
		(types
		    (list (mercury-make-keywords-regexp mercury-types-i t) 0 'font-lock-type-face))
		(modes
		    (list
			    (mercury-make-keywords-regexp mercury-mode-specificators-i t)
				0 'font-lock-constant-face))
		(directives
		    (list
			    (mercury-make-keywords-regexp mercury-directives-i t)
				0 'font-lock-warning-face))
    ) ; =========================================================================
        ;; Make font lock list
        (delq nil
            (cond
                ((eq major-mode 'mercury-MAJOR-mode)
                    (list
                        head-predicates quoted_atom string_operator string variables
						important-elements important-elements-1 predspecs keywords
						types modes directives))
				((eq major-mode 'mercury-inferior-mode)
				    (list
					    prompt error-message-identifier error-whole-messages
					    error-warning-messages warning-messages predspecs))
				((eq major-mode 'compilation-mode)
				    (list
					    predspecs))
			)
		)
    ) ; =========================================================================
)




;;-------------------------------------------------------------------
;; User configurable variables
;;-------------------------------------------------------------------
;; General configuration
(defcustom mercury-system nil
    "*Mercury interpreter/compiler used.
The value of this variable is nil or a symbol.
If it is a symbol, it determines default values of other configuration
variables with respect to properties of the specified Mercury
interpreter/compiler.

Currently recognized symbol values are:
mercury - Mercury"
    :group 'mercury
    :type '(choice (const :tag "Default" :value nil))
)

(make-variable-buffer-local 'mercury-system)

;; NB: This alist can not be processed in mercury-mode-variables to
;; create a mercury-system-version-i variable since it is needed
;; prior to the call to mercury-mode-variables.
(defcustom mercury-system-version
    '( (mercury  (0 . 13)) )
    "*Alist of Mercury system versions.
The version numbers are of the format (Major . Minor)."
)
;; Indentation
(defcustom mercury-indent-width tab-width
    "*The indentation width used by the editing buffer."
    :group 'mercury-indentation
    :type 'integer)  
(defcustom mercury-align-comments-flag t
    "*Non-nil means automatically align comments when indenting."
    :group 'mercury-indentation
    :type 'boolean)
(defcustom mercury-indent-mline-comments-flag t
    "*Non-nil means indent contents of /* */ comments.
Otherwise leave such lines as they are."
    :group 'mercury-indentation
    :type 'boolean)
(defcustom mercury-object-end-to-0-flag t
    "*Non-nil means indent closing '}' in SICStus object definitions to level 0.
Otherwise indent to `mercury-indent-width'."
	:group 'mercury-indentation
	:type 'boolean)
(defcustom mercury-left-indent-regexp "\\(;\\|\\*?->\\)"
    "*Regexp for character sequences after which next line is indented.
Next line after such a regexp is indented to the opening paranthesis level."
    :group 'mercury-indentation
    :type 'regexp)
(defcustom mercury-paren-indent-p nil
    "*If non-nil, increase indentation for parenthesis expressions.
The second and subsequent line in a parenthesis expression other than
a compound term can either be indented `mercury-paren-indent' to the
right (if this variable is non-nil) or in the same way as for compound
terms (if this variable is nil, default)."
	:group 'mercury-indentation
	:type 'boolean)
(defcustom mercury-paren-indent 4
    "*The indentation increase for parenthesis expressions.
Only used in ( If -> Then ; Else) and ( Disj1 ; Disj2 ) style expressions."
    :group 'mercury-indentation
    :type 'integer)
(defcustom mercury-parse-mode 'beg-of-clause
    "*The parse mode used (decides from which point parsing is done).
Legal values:
'beg-of-line   - starts parsing at the beginning of a line, unless the
                 previous line ends with a backslash. Fast, but has
                 problems detecting multiline /* */ comments.
'beg-of-clause - starts parsing at the beginning of the current clause.
                 Slow, but copes better with /* */ comments."
    :group 'mercury-indentation
    :type '(choice (const :value beg-of-line) (const :value beg-of-clause)))
;; Font locking
(defcustom mercury-keywords
    '((mercury (
        "all" "else" "end_module" "equality" "external" "fail" "func" "if"
		"implementation" "import_module" "include_module" "inst" "instance"
		"interface" "mode" "module" "not" "pragma" "pred" "some" "then" "true"
		"type" "typeclass" "use_module" "where" "main" 
		;; Nick
		"lambda" "and" "or" "rem" "fact_table" "type_spec" "promise_pure"
		"impure" "inline" "foreign_proc" "foreign_decl" "throw" "div" "mod"
		"domain_checks" "promise" "obsolete" "will_not_call_mercury" "thread_safe"
		"rule" "when" "semipure" "aditi_bottom_up" "aditi_top_down" "try" "try_io"
		"try_store" "try_all" "incremental_try_all" "rethrow" "finally"
		"throw_if_near_stack_limits" "foreign_code" "foreign_type" 
		"may_call_mercury" "export" "error" "solver" "semidet_succeed"
		"c_header_code" "c_code" "not_thread_safe" "mutable" "untrailed"
		"representation" "initialisation" "will_not_modify_trail"
		"initialization" "constraint_store" "comparison" "trailed" "untrailed"
		"attach_to_io_state" "foreign_name" "memo" "call"
		
		"no" "yes"
	)))
    "*Alist of Mercury keywords which is used for font locking of directives."
	:group 'mercury-font-lock
	:type 'sexp)
;;  Modules:
(defcustom mercury-types
    '((mercury (
        ;;;   library:
		"array" "array2d" "assoc_list" "bag" "benchmarking" "bimap" "bintree"
		"bintree_set" "bitmap" "bool" "bt_array" "builtin" "char" "construct"
		"cord" "counter" "deconstruct" "dir" "enum" "eqvclass" "exception"
		"float" "gc" "getopt" "getopt_io" "graph" "group" "hash_table"
		"injection" "int" "integer" "io" "lexer" "library" "list" "map" "math"
		"maybe" "multi_map" "ops" "pair" "parser" "pprint" "pqueue" "prolog"
		"queue" "random" "rational" "rbtree" "relation" "require" "set"
		"set_bbbtree" "set_ctree234" "set_ordlist" "set_tree234" "set_unordlist"
		"solutions" "sparse_bitset" "stack" "std_util" "store" "string" 
		"svarray" "svbag" "svmulti-map" "svqueue" "svrelation" "svset" 
		"svvarset" "term" "term_io" "term_to_xml" "time" "tree234" "type_desc"
		"unit" "univ" "varset" "version_array" "version_array2d" "version_bitmap"
		"version_hash_table" "version_store" "version_types"
	    ;;;   extras:
		"cgi" "html" "http" "odbc" "parsing" "stream" "tcl" "tcp" "tk" 
		"unicode" 
		"qcheck" "rnd"  ;;  quickcheck
		"xml"  ;; "xml.cat" "xml.encoding" "xml.parse" "xml.ns"
		
		"c_pointer"
    )))
    "*Alist of Mercury package specificators used by font locking."
	:group 'mercury-font-lock
	:type 'sexp
)
(defcustom mercury-mode-specificators
    '((mercury (
        ;;; insts
		"array_skel" "uniq_array" "unique" "bound" "free" "ground" "uniq_array_skel" 
		"mostly_uniq_array" "mostly_unique" "mostly_uniq_array_skel" "uniq_bitmap"
		"dead" "clobbered" "mostly_dead" "mostly_clobbered" "any" "old" "new"
		"unify" "compare" "comparison_pred" "comparison_func" "do_not_allow"
		"canonicalize" "include_details_cc" "option_ops" "option_ops_track"
		"hash_pred" "list_skel" "non_empty_list" "xml_doc" "non_embedded_dtd"
		"element_pred"
	    ;;; modes
		"array2d_di" "array2d_ui" "array2d_uo" "array_di" "array_ui" "array_uo" 
		"array_mdi" "array_mui" "array_muo" "bitmap_ui"  "bitmap_di"  "bitmap_uo"
		"unused" "output" "input" "in"  "out" "di" "mdi" "uo" "ui" "mui" "muo" "mdi"
		"oa" "ia" "hash_table_di" "hash_table_ui" "hash_table_uo" "in_list_skel" 
		"out_list_skel" "list_skel_out" "list_skel_in" "canonicalize_or_do_not_allow"
		"foldl_pred" "cannot_fail" "nil_or_singleton_list" "try_all_det"
		"try_all_semidet" "try_all_multi" "try_all_nondet" 

		"pdi" "puo"
	)))
	"*Alist of Mercury inst and mode specificators used by font locking."
	:group 'mercury-font-lock
	:type 'sexp
)
(defcustom mercury-determinism-specificators
    '((mercury (
        "cc_multi" "cc_nondet" "det" "erroneous" "failure" "multi" "nondet" "semidet" 
		;; Nick
		"is"
    )))
	"*Alist of Mercury determinism specificators used by font locking."
	:group 'mercury-font-lock
	:type 'sexp)
(defcustom mercury-directives
    '((mercury ("^#[0-9]+")))
	"*Alist of Mercury source code directives used by font locking."
	:group 'mercury-font-lock
	:type 'sexp)
;; Keyboard
(defcustom mercury-electric-newline-flag t
    "*Non-nil means automatically indent the next line when the user types RET."
	:group 'mercury-keyboard
	:type 'boolean)
(defcustom mercury-hungry-delete-key-flag nil
    "*Non-nil means delete key consumes all preceding spaces."
	:group 'mercury-keyboard
	:type 'boolean)
(defcustom mercury-electric-dot-flag nil
    "*Non-nil means make dot key electric.
Electric dot appends newline or inserts head of a new clause.
If dot is pressed at the end of a line where at least one white space
precedes the point, it inserts a recursive call to the current predicate.
If dot is pressed at the beginning of an empty line, it inserts the head
of a new clause for the current predicate. It does not apply in strings
and comments. 
It does not apply in strings and comments."
	:group 'mercury-keyboard
	:type 'boolean)
(defcustom mercury-electric-underscore-flag nil
    "*Non-nil means make underscore key electric.
Electric underscore replaces the current variable with underscore.
If underscore is pressed not on a variable then it behaves as usual."
	:group 'mercury-keyboard
	:type 'boolean)
(defcustom mercury-electric-tab-flag nil
    "*Non-nil means make TAB key electric.
Electric TAB inserts spaces after parentheses, ->, and ;
in ( If -> Then ; Else) and ( Disj1 ; Disj2 ) style expressions."
	:group 'mercury-keyboard
	:type 'boolean)
;; Inferior mode
(defcustom mercury-compile-string
    '((mercury "mmake "))
	"*Alist of strings and lists defining predicate for recompilation.

Some parts of the string are replaced:
`%f' by the name of the compiled file (can be a temporary file)
`%b' by the file name of the buffer to compile
`%m' by the module name and name of the compiled file separated by colon
`%l' by the line offset into the file. This is 0 unless compiling a
     region of a buffer, in which case it is the number of lines before
     the region.

It is an argument to the `compile' function."
	:group 'mercury-inferior
	:type 'sexp)
(defcustom mercury-eof-string "end_of_file.\n"
    "*Alist of strings that represent end of file for mercury.
nil means send actual operating system end of file."
	:group 'mercury-inferior
	:type 'sexp)
(defcustom mercury-use-standard-consult-compile-method-flag t
    "*Non-nil means use the standard compilation method.
Otherwise the new compilation method will be used. This
utilises a special compilation buffer with the associated
features such as parsing of error messages and automatically
jumping to the source code responsible for the error.

Warning: the new method is so far only experimental and
does contain bugs. The recommended setting for the novice user
is non-nil for this variable."
	:group 'mercury-inferior
	:type 'boolean)
;; Miscellaneous
(defcustom mercury-use-mercury-tokenizer-flag t
    "*Non-nil means use the internal mercury tokenizer for indentation etc.
Otherwise use `parse-partial-sexp' which is faster but sometimes incorrect."
	:group 'mercury-other
	:type 'boolean)
(defcustom mercury-imenu-flag t
    "*Non-nil means add a clause index menu for all mercury files."
	:group 'mercury-other
	:type 'boolean)
(defcustom mercury-imenu-max-lines 3000
    "*The maximum number of lines of the file for imenu to be enabled.
Relevant only when `mercury-imenu-flag' is non-nil."
	:group 'mercury-other
	:type 'integer)
(defcustom mercury-info-predicate-index
    "(sicstus)Predicate Index"
	"*The info node for the SICStus predicate index."
	:group 'mercury-other
	:type 'string)
(defcustom mercury-underscore-wordchar-flag t
    "*Non-nil means underscore (_) is a word-constituent character."
	:group 'mercury-other
	:type 'boolean)
(defcustom mercury-char-quote-workaround t
    "*If non-nil, declare 0 as a quote character so that 0'<char> does not break syntax highlighting.
This is really kludgy but I have not found any better way of handling it."
	:group 'mercury-other
	:type 'boolean)




;;-------------------------------------------------------------------
;; Internal variables
;;-------------------------------------------------------------------
(defvar mercury-emacs 
    (if (string-match "XEmacs\\|Lucid" emacs-version)
        'xemacs
        'gnuemacs)
	"The variant of Emacs we're running.
Valid values are 'gnuemacs and 'xemacs.")
(defvar mercury-known-systems '(eclipse mercury sicstus swi gnu))

;(defvar mercury-temp-filename "")   ; Later set by `mercury-temporary-file'

(defvar mercury-mode-syntax-table nil)
(defvar mercury-mode-abbrev-table nil)
(defvar mercury-mode-map nil)
(defvar mercury-upper-case-string ""
    "A string containing all upper case characters.
Set by mercury-build-case-strings.")
(defvar mercury-lower-case-string ""
    "A string containing all lower case characters.
Set by mercury-build-case-strings.")
(defvar mercury-atom-char-regexp ""
    "Set by mercury-set-atom-regexps.")
    ;; "Regexp specifying characters which constitute atoms without quoting.")
(defvar mercury-compilation-buffer "*mercury-compilation*"
  "Name of the output buffer for Mercury compilation/consulting.")
(defvar mercury-keywords-i nil)
(defvar mercury-types-i nil)
(defvar mercury-mode-specificators-i nil)
(defvar mercury-determinism-specificators-i nil)
(defvar mercury-directives-i nil)
(defvar mercury-compile-string-i nil)
(defvar mercury-eof-string-i nil)
(defvar mercury-mode-hook nil
    "List of functions to call after the Mercury mode has initialised.")
(defvar mercury-inferior-mode-map nil)
(defvar mercury-inferior-mode-hook nil
    "List of functions to call after the inferior mercury mode has initialised.")
;;; The rest of this defvars is experimental code!
;; Global variables for process filter function
(defvar mercury-process-flag nil
    "Non-nil means that a mercury task (i.e. a consultation or compilation job) 
is running.")
(defvar mercury-consult-compile-output ""
    "Hold the unprocessed output from the current mercury task.")
(defvar mercury-consult-compile-first-line 1
    "The number of the first line of the file to consult/compile.
Used for temporary files.")
(defvar mercury-consult-compile-file nil
    "The file to compile/consult (can be a temporary file).")
(defvar mercury-consult-compile-real-file nil
    "The file name of the buffer to compile/consult.")
;;-------------------------------------------------------------------
;; Online help
;;-------------------------------------------------------------------
(defvar mercury-help-function
    '(
	    (mercury nil)
        ;;;;(eclipse mercury-help-online)
		;; (sicstus mercury-help-info)
        ;;;;(sicstus mercury-find-documentation)
        ;;;;(swi mercury-help-online)
        ;;;;(t mercury-help-online)
	)
	"Alist for the name of the function for finding help on a predicate."
)


(defconst mercury-atom-regexp ""
    "Set by mercury-set-atom-regexps.")
(defconst mercury-dot-name-regexp "^\\[a-z][a-zA-Z0-9_]\\)"
    "Set by mercury-set-atom-regexps.")
(defconst mercury-left-paren "[[({]" 
    "The characters used as left parentheses for the indentation code.")
(defconst mercury-right-paren "[])}]"
    "The characters used as right parentheses for the indentation code.")
(defconst mercury-quoted-atom-regexp "\\(^\\|[^0-9]\\)\\('\\([^\n']\\|\\\\'\\)*'\\)"
    "Regexp matching a quoted atom.")
(defconst mercury-string-operator-regexp "\\(`\\([^\n`]\\|\\\\`\\)*`\\)"
    "Regexp matching a string.")
;;  NICK: \(\" \([^\n\"]\|\\\"\)* \"\)
(defconst mercury-string-regexp "\\(\"\\([^\n\"]\\|\\\\\"\\)*\"\\)"
    "Regexp matching a string.")
(defconst mercury-head-delimiter "\\(:-\\|\\+:\\|-:\\|\\+\\?\\|-\\?\\|-->\\)"
    "A regexp for matching on the end delimiter of a head (e.g. \":-\").")
(defconst mercury-tokenize-searchkey
    (concat 
	    "[0-9]+'"
		"\\|"
		"['\"]"
		"\\|"
		mercury-left-paren
		"\\|"
		mercury-right-paren
		"\\|"
		"%"
		"\\|"
		"/\\*")
)


;;-------------------------------------------------------------------
;; Mercury mode
;;-------------------------------------------------------------------
;; Example: (mercury-atleast-version '(3 . 6))
(defun mercury-atleast-version (version)
    "Return t if the version of the current Mercury system is VERSION or later.
VERSION is of the format (Major . Minor)"
	;; Version.major < major or
	;; Version.major = major and Version.minor <= minor
	(let* ( ;====================================================================
		(thisversion (mercury-find-value-by-system mercury-system-version))
        (thismajor (car thisversion))
        (thisminor (cdr thisversion))
	) ;==========================================================================
	    (or 
		    (< (car version) thismajor)
			(and (= (car version) thismajor) (<= (cdr version) thisminor)))
    ) ;==========================================================================
)


(if mercury-mode-syntax-table
    ()
    (let ( ;====================================================================
		(table (make-syntax-table))
    ) ;=========================================================================
        (if mercury-underscore-wordchar-flag
            (modify-syntax-entry ?_ "w" table)
            (modify-syntax-entry ?_ "_" table))
		(modify-syntax-entry ?+ "." table)
		(modify-syntax-entry ?- "." table)
		(modify-syntax-entry ?= "." table)
		(modify-syntax-entry ?< "." table)
		(modify-syntax-entry ?> "." table)
		(modify-syntax-entry ?| "." table)
		(modify-syntax-entry ?\' "\"" table)
		;; -------------------------------------------------------------
		;; Macht Gnsefsschen hinter 0 ungltig - zu hoher Preis...
		;; TODO
		;; Any better way to handle the 0'<char> construct?!?
		;NICK;(when mercury-char-quote-workaround
        ;NICK;  (modify-syntax-entry ?0 "\\" table))
		(modify-syntax-entry ?% "<" table)
		(modify-syntax-entry ?\n ">" table)
		(if (eq mercury-emacs 'xemacs)
			(progn
			    (modify-syntax-entry ?* ". 67" table)
				(modify-syntax-entry ?/ ". 58" table))
		    ;; Emacs wants to see this it seems:
		    (modify-syntax-entry ?* ". 23b" table)
			(modify-syntax-entry ?/ ". 14" table))
		(setq mercury-mode-syntax-table table)
	) ;==========================================================================
)

(define-abbrev-table 'mercury-mode-abbrev-table ())

(defun mercury-find-value-by-system (alist)
    "Get value from ALIST according to `mercury-system'."
	(if (listp alist)
        (let ( ;=================================================================
		    result
            id
	    ) ;======================================================================
            (while alist
			    (setq id (car (car alist)))
			    (if 
				    (or 
				        (eq id mercury-system)
						(eq id t)
						(and (listp id) (eval id)))
					(progn
					    (setq result (car (cdr (car alist))))
					    (if 
						    (and  
						        (listp result)  
							    (eq (car result) 'eval))
							(setq result (eval (car (cdr result)))))
						(setq alist nil))
				    (setq alist (cdr alist)))
			)
			result
		) ;======================================================================
	    alist
	)
)

(defun mercury-mode-variables ()
    "Set some common variables to Mercury code specific values."
	(make-local-variable 'adaptive-fill-mode)  ;;  adaptive-fill-mode
	(setq adaptive-fill-mode nil)
	(make-local-variable 'comment-end)  ;;  comment-end
	(setq comment-end "")
	(make-local-variable 'comment-indent-function)  ;;  comment-indent-function
	(setq comment-indent-function 'mercury-comment-indent)
	(make-local-variable 'comment-start)  ;;  comment-start
	(setq comment-start "%")
	;; This complex regexp makes sure that comments cannot start
	;; inside quoted atoms or strings
	(make-local-variable 'comment-start-skip)  ;;  comment-start-skip
	(setq comment-start-skip 
        (format 
		    "^\\(\\(%s\\|%s\\|[^\n\'\"%%]\\)*\\)\\(/\\*+ *\\|%%+ *\\)" 
			mercury-quoted-atom-regexp mercury-string-regexp))
	(make-local-variable 'compile-command)  ;;  compile-command
	(setq compile-command mercury-compile-string-i)
	(make-local-variable 'font-lock-defaults)  ;;  font-lock-defaults
	(setq font-lock-defaults
        '(mercury-font-lock-keywords nil nil ((?_ . "w"))))
	(make-local-variable 'indent-line-function)  ;;  indent-line-function
	(setq indent-line-function 'mercury-indent-line)
	(make-local-variable 'paragraph-ignore-fill-prefix)  ;;  paragraph-ignore-fill-prefix
	(setq paragraph-ignore-fill-prefix t)
	(make-local-variable 'paragraph-separate)  ;;  paragraph-separate
	(setq paragraph-separate paragraph-start)
	(make-local-variable 'paragraph-start)  ;;  paragraph-start
	(setq paragraph-start (concat "[ \t]*$\\|" page-delimiter)) ;'%%..'
	(make-local-variable 'parens-require-spaces)  ;;  parens-require-spaces
	(setq parens-require-spaces nil)

	(setq local-abbrev-table mercury-mode-abbrev-table)

	(make-local-variable 'comment-column)
	;; Initialize Mercury system specific variables
	(let ( ;===================================================================== 0000bf/00ffff
		(vars '(
            mercury-keywords 
			mercury-types 
			mercury-mode-specificators
			mercury-determinism-specificators 
			mercury-directives
			mercury-compile-string 
			mercury-eof-string
		))
	) ;========================================================================== ffff00/00007f
	    (while vars
		    (set (intern (concat (symbol-name (car vars)) "-i"))
				(mercury-find-value-by-system (eval (car vars))))
			(setq vars (cdr vars))
		)
	) ;==========================================================================
)

(defun mercury-mode-keybindings-common (map)
    "Define keybindings common to both Mercury modes in MAP."
	(define-key map "\C-c?" 'mercury-help-on-predicate)
)

(defun mercury-mode-keybindings-edit (map)
    "Define keybindings for Mercury mode in MAP."
	(define-key map "\M-a" 'mercury-beginning-of-clause)
	(define-key map "\M-e" 'mercury-end-of-clause)
	(define-key map "\C-\M-a" 'mercury-beginning-of-predicate)
	(define-key map "\C-\M-e" 'mercury-end-of-predicate)
	(define-key map "\M-\C-c" 'mercury-mark-clause)
	(define-key map "\M-\C-h" 'mercury-mark-predicate)
	(define-key map "\M-\C-n" 'mercury-forward-list)
	(define-key map "\M-\C-p" 'mercury-backward-list)
	(define-key map "\C-c\C-n" 'mercury-insert-predicate-template)
	(define-key map "\C-c\C-s" 'mercury-insert-predspec)
	(define-key map "\M-\r" 'mercury-insert-next-clause)
	(define-key map "\C-c\C-va" 'mercury-variables-to-anonymous)
	(define-key map "\C-c\C-v\C-s" 'mercury-view-predspec)
	(define-key map [Backspace] 'mercury-electric-delete)
	(define-key map "." 'mercury-electric-dot)
	(define-key map "_" 'mercury-electric-underscore)
	(if mercury-electric-newline-flag 
        (define-key map "\r" 'newline-and-indent))
)

(defun mercury-mode-keybindings-inferior (map)
    "Define keybindings for inferior Mercury mode in MAP."
    ;; No inferior mode specific keybindings now.
)

(if mercury-mode-map
    ()
    (setq mercury-mode-map (make-sparse-keymap))
	(mercury-mode-keybindings-common mercury-mode-map)
	(mercury-mode-keybindings-edit mercury-mode-map)
	;; System dependent keymaps for system dependent menus
	(let (
		(systems mercury-known-systems)
    )
        (while systems
		    (set (intern (concat "mercury-mode-map-" (symbol-name (car systems))))
                ;(cons 'keymap mercury-mode-map)
				mercury-mode-map)
			(setq systems (cdr systems))
		)
	)
)


;;;###autoload
(defun mercury-MAJOR-mode (&optional system)
    "Major mode for editing Mercury code.

Blank lines and `%%...' separate paragraphs.  `%'s starts a comment
line and comments can also be enclosed in /* ... */.

If an optional argument SYSTEM is non-nil, set up mode for the given system.

To find out what version of Mercury mode you are running, enter
`\\[mercury-mode-version]'.

Commands:
\\{mercury-mode-map}
Entry to this mode calls the value of `mercury-mode-hook'
if that value is non-nil."
	(interactive)
	(kill-all-local-variables)
	(if system 
		(setq mercury-system system)
	)
	(use-local-map
	    (if mercury-system 
            ;; ### Looks like it works under XEmacs as well...
            ;; (and mercury-system
            ;;     (not (eq mercury-emacs 'xemacs)))
            (eval (intern (concat "mercury-mode-map-" (symbol-name mercury-system))))
		    mercury-mode-map
	    )
    )
	(setq major-mode 'mercury-MAJOR-mode)
	(setq mode-name "Mercury")
	(set-syntax-table mercury-mode-syntax-table)
	(mercury-mode-variables)
	(mercury-build-case-strings)
	(mercury-set-atom-regexps)
	;; imenu entry moved to the appropriate hook for consistency  
	(run-hooks 'mercury-mode-hook)
)

;;;###autoload
(defun mercury-mode ()
    "Major mode for editing Mercury programs.
Actually this is just customized `mercury-MAJOR-mode'."
	(interactive)
	(mercury-MAJOR-mode 'mercury)
)


;;-------------------------------------------------------------------
;; Inferior mercury mode
;;-------------------------------------------------------------------
(defun mercury-inferior-mode ()
    "Major mode for interacting with an inferior Mercury process.

The following commands are available:
\\{mercury-inferior-mode-map}

Entry to this mode calls the value of `mercury-mode-hook' with no arguments,
if that value is non-nil.  Likewise with the value of `comint-mode-hook'.
`mercury-mode-hook' is called after `comint-mode-hook'.

You can send text to the inferior Mercury from other buffers
using the commands `send-region' and `send-string'.

Commands:
Tab indents for Mercury; with argument, shifts rest
 of expression rigidly with the current line.
Paragraphs are separated only by blank lines and '%%'. '%'s start comments.

Return at end of buffer sends line as input.
Return not at end copies rest of line to end and sends it.
\\[comint-delchar-or-maybe-eof] sends end-of-file as input.
\\[comint-kill-input] and \\[backward-kill-word] are kill commands,
imitating normal Unix input editing.
\\[comint-interrupt-subjob] interrupts the shell or its current subjob if any.
\\[comint-stop-subjob] stops, likewise.
\\[comint-quit-subjob] sends quit signal, likewise.

To find out what version of Mercury mode you are running, enter
`\\[mercury-mode-version]'."
	(interactive)
	(cond 
	    ((not (eq major-mode 'mercury-inferior-mode))
            (kill-all-local-variables)
			(comint-mode)
			(setq comint-input-filter 'mercury-input-filter)
			(setq major-mode 'mercury-inferior-mode)
			(setq mode-name "Inferior Mercury")
			(setq mode-line-process '(": %s"))
			(mercury-mode-variables)
			(if mercury-inferior-mode-map
                ()
                (setq mercury-inferior-mode-map (copy-keymap comint-mode-map))
				(mercury-mode-keybindings-common mercury-inferior-mode-map)
				(mercury-mode-keybindings-inferior mercury-inferior-mode-map))
			(use-local-map mercury-inferior-mode-map)
            ;;;;(setq comint-prompt-regexp mercury-prompt-regexp-i) <======================
			(make-variable-buffer-local 'shell-dirstack-query)
			(setq shell-dirstack-query "pwd.")
			(run-hooks 'mercury-inferior-mode-hook))
	)
)

(defun mercury-input-filter (str)
    (cond ((string-match "\\`\\s *\\'" str) nil) ;whitespace
        ((not (eq major-mode 'mercury-inferior-mode)) t)
        ((= (length str) 1) nil)        ;one character
        ((string-match "\\`[rf] *[0-9]*\\'" str) nil) ;r(edo) or f(ail)
        (t t)))

(defun mercury-process-insert-string (process string)
    "Insert STRING into inferior Mercury buffer running PROCESS."
	;; Copied from elisp manual, greek to me
	(let (
		(buf (current-buffer))
	)
	    (unwind-protect
			(let (
				moving
			)
			    (set-buffer (process-buffer process))
				(setq moving (= (point) (process-mark process)))
				(save-excursion
				    ;; Insert the text, moving the process-marker.
				    (goto-char (process-mark process))
					(insert string)
					(set-marker (process-mark process) (point))
				)
				(if moving 
					(goto-char (process-mark process))
				)
			)
		    (set-buffer buf)
		)
	)
)


;;------------------------------------------------------------
;; Old consulting and compiling functions
;;------------------------------------------------------------

(defun mercury-old-process-region (compilep start end) 
    "Process the region limited by START and END positions.
If COMPILEP is non-nil then use compilation, otherwise consulting."
	;(let ((tmpfile mercury-temp-filename)
	(let ( ;=====================================================================
		(tmpfile (mercury-bsts (mercury-temporary-file)))
;;;; Sinnvoll, wo Mercury ohne Interpreter??
;;  ??????????????????????????????????????????????????????????????????????????
        ;(process (get-process "mercury"))
		(first-line 
		    (1+ (count-lines 
                (point-min) 
				(save-excursion (goto-char start) (point))
			))
		)
	) ;==========================================================================
        (write-region start end tmpfile)
		(process-send-string "mercury" 
		    (mercury-build-mercury-command
                compilep 
				tmpfile 
				(mercury-bsts buffer-file-name)
				first-line
			)
		)
		(mercury-goto-mercury-process-buffer)
	) ;==========================================================================
)

(defun mercury-old-process-predicate (compilep)
    "Process the predicate around point.
If COMPILEP is non-nil then use compilation, otherwise consulting."
	(mercury-old-process-region
	    compilep (mercury-pred-start) (mercury-pred-end))
)

(defun mercury-old-process-buffer (compilep)
    "Process the entire buffer.
If COMPILEP is non-nil then use compilation, otherwise consulting."
    (mercury-old-process-region 
	    compilep (point-min) (point-max))
)

(defun mercury-old-process-file (compilep)
    "Process the file of the current buffer.
If COMPILEP is non-nil then use compilation, otherwise consulting."
    (save-some-buffers)
    (let ( ;=====================================================================
		(filename (mercury-bsts buffer-file-name))
	) ;==========================================================================
	    (process-send-string "mercury" 
			(mercury-build-mercury-command 
                compilep filename filename))
		(mercury-goto-mercury-process-buffer)
	) ;==========================================================================
)


;;------------------------------------------------------------
;; Consulting and compiling
;;------------------------------------------------------------
;;; Interactive interface functions, used by both the standard
;;; and the experimental consultation and compilation functions

(defun mercury-compile-file ()
    "Compile file of current buffer."
    (interactive)
    (if mercury-use-standard-consult-compile-method-flag
        (mercury-old-process-file t)
	  (mercury-consult-compile-file t)
    )
)

(defun mercury-compile-buffer ()
    "Compile buffer."
    (interactive)
    (if mercury-use-standard-consult-compile-method-flag
		(mercury-old-process-buffer t)
	    (mercury-consult-compile-buffer t)
	)
)

(defun mercury-compile-region (beg end)
    "Compile region between BEG and END."
	(interactive "r")
	(if mercury-use-standard-consult-compile-method-flag
		(mercury-old-process-region t beg end)
	    (mercury-consult-compile-region t beg end)
	)
)

(defun mercury-compile-predicate ()
    "Compile the predicate around current point."
	(interactive)
	(if mercury-use-standard-consult-compile-method-flag
		(mercury-old-process-predicate t)
	    (mercury-consult-compile-predicate t)
	)
)

(defun mercury-buffer-module ()
    "Select Mercury module name appropriate for current buffer.
Bases decision on buffer contents (-*- line)."
	;; Look for -*- ... module: MODULENAME; ... -*-
	(let ( ;=====================================================================
		beg 
		end
	) ;==========================================================================
	    (save-excursion
		    (goto-char (point-min))
			(skip-chars-forward " \t")
			(and 
			    (search-forward "-*-" (save-excursion (end-of-line) (point)) t)
				(progn
				    (skip-chars-forward " \t")
					(setq beg (point))
					(search-forward "-*-" (save-excursion (end-of-line) (point)) t)
				)
				(progn
				    (forward-char -3)
					(skip-chars-backward " \t")
					(setq end (point))
					(goto-char beg)
					(and 
					    (let ( ;=====================================================================
							(case-fold-search t)
						) ;==========================================================================
						    (search-forward "module:" end t)
						) ;==========================================================================
						(progn
						    (skip-chars-forward " \t")
							(setq beg (point))
							(if (search-forward ";" end t)
								(forward-char -1)
							    (goto-char end)
							)
							(skip-chars-backward " \t")
							(buffer-substring beg (point))
						)
					)
				)
			)
		)
	) ;==========================================================================
)

(defun mercury-build-mercury-command (compilep file buffername 
                                    &optional first-line)
    "Make Mercury command for FILE compilation/consulting.
If COMPILEP is non-nil, consider compilation, otherwise consulting."
    (let* ( ;====================================================================
		(compile-string mercury-compile-string-i)
	    ;;;; `compilep' auch weg?? <======================================
        ;;;;(if compilep 
	    ;;;;    mercury-compile-string-i 
	    ;;;;    mercury-consult-string-i
	    ;;;;)
		(module (mercury-buffer-module))
		(file-name (concat "'" file "'"))
		(module-name 
		    (if module (concat "'" module "'")))
        (module-file 
		    (if module
                (concat module-name ":" file-name)
                file-name
			))
		strbeg 
		strend
		(lineoffset 
		    (if first-line
				(- first-line 1)
			    0
			))
	) ;==========================================================================
	    ;; Assure that there is a buffer name
        (if (not buffername)
			(error "The buffer is not saved"))
		(if (not (string-match "^'.*'$" buffername)) ; Add quotes
			(setq buffername (concat "'" buffername "'")))
		(while (string-match "%m" compile-string)
		    (setq strbeg (substring compile-string 0 (match-beginning 0)))
			(setq strend (substring compile-string (match-end 0)))
			(setq compile-string (concat strbeg module-file strend)))
		(while (string-match "%f" compile-string)
		    (setq strbeg (substring compile-string 0 (match-beginning 0)))
			(setq strend (substring compile-string (match-end 0)))
			(setq compile-string (concat strbeg file-name strend)))
		(while (string-match "%b" compile-string)
		    (setq strbeg (substring compile-string 0 (match-beginning 0)))
			(setq strend (substring compile-string (match-end 0)))
			(setq compile-string (concat strbeg buffername strend)))
		(while (string-match "%l" compile-string)
		    (setq strbeg (substring compile-string 0 (match-beginning 0)))
			(setq strend (substring compile-string (match-end 0)))
			(setq compile-string (concat strbeg (format "%d" lineoffset) strend)))
		(concat compile-string "\n")
	) ;==========================================================================
)

;;; The rest of this page is experimental code!


(defun mercury-consult-compile-filter (process output)
    "Filter function for Mercury compilation PROCESS.
Argument OUTPUT is a name of the output file."
    ;;(message "start")
    (setq mercury-consult-compile-output
        (concat mercury-consult-compile-output output))
	;;(message "pccf1: %s" mercury-consult-compile-output)
	;; Iterate through the lines of mercury-consult-compile-output
	(let ( ;=====================================================================
		outputtype
	) ;==========================================================================
	    (while 
			(and 
			    mercury-process-flag
                (or
                    ;; Trace question
                    (progn 
                        (setq outputtype 'trace)
                        (and 
						    (eq mercury-system 'sicstus)
                            (string-match
							    "^[ \t]*[0-9]+[ \t]*[0-9]+[ \t]*Call:.*? "
								mercury-consult-compile-output))
					)
					;; Match anything
					(progn 
					    (setq outputtype 'normal)
						(string-match "^.*\n" mercury-consult-compile-output)
					))
			)
		    ;;(message "outputtype: %s" outputtype)
		    (setq output (match-string 0 mercury-consult-compile-output))
			;; remove the text in output from mercury-consult-compile-output
			(setq mercury-consult-compile-output
				(substring mercury-consult-compile-output (length output)))
			;;(message "pccf2: %s" mercury-consult-compile-output)

			;; Write the output in the *mercury-compilation* buffer
			(insert output))
	) ;==========================================================================
	;; If the prompt is visible, then the task is finished
    ;;;;(if (string-match mercury-prompt-regexp-i mercury-consult-compile-output) <==================
    ;;;;    (setq mercury-process-flag nil)
    ;;;;)
)

(defun mercury-consult-compile-file (compilep)
    "Consult/compile file of current buffer.
If COMPILEP is non-nil, compile, otherwise consult."
    (let ( ;=====================================================================
		(file buffer-file-name)
	) ;==========================================================================
	    (if file
			(progn
			    (save-some-buffers)
				(mercury-consult-compile compilep file)
			)
		    (mercury-consult-compile-region compilep (point-min) (point-max)))
	) ;==========================================================================
)

(defun mercury-consult-compile-buffer (compilep)
    "Consult/compile current buffer.
If COMPILEP is non-nil, compile, otherwise consult."
    (mercury-consult-compile-region compilep (point-min) (point-max))
)

(defun mercury-consult-compile-region (compilep beg end)
    "Consult/compile region between BEG and END.
If COMPILEP is non-nil, compile, otherwise consult."
    ;(let ((file mercury-temp-filename)
    (let ( ;=====================================================================
		(file 
		    (mercury-bsts (mercury-temporary-file)))
        (lines (count-lines 1 beg))
	) ;==========================================================================
	    (write-region beg end file nil 'no-message)
		(write-region "\n" nil file t 'no-message)
		(mercury-consult-compile 
		    compilep 
			file
			(if (looking-at "^") 
				(1+ lines) 
			    lines))
		(delete-file file)
	)
)

(defun mercury-consult-compile-predicate (compilep)
    "Consult/compile the predicate around current point.
If COMPILEP is non-nil, compile, otherwise consult."
    (mercury-consult-compile-region
	    compilep (mercury-pred-start) (mercury-pred-end))
)

;;-------------------------------------------------------------------
;; Font-lock stuff
;;-------------------------------------------------------------------

;; Auxilliary functions
(defun mercury-make-keywords-regexp (keywords &optional protect)
    "Create regexp from the list of strings KEYWORDS.
If PROTECT is non-nil, surround the result regexp by word breaks."
    (let ( ;=====================================================================
		(regexp
            (if (fboundp 'regexp-opt)
                ;; Emacs 20
                ;; Avoid compile warnings under earlier versions by using eval
				(eval '(regexp-opt keywords))
			    ;; Older Emacsen
			    (concat (mapconcat 'regexp-quote keywords "\\|"))))
	) ;==========================================================================
	    (if protect
			(concat "\\<\\(" regexp "\\)\\>")
		    regexp)
	) ;==========================================================================
)

(defun mercury-font-lock-object-matcher (bound)
    "Find SICStus objects method name for font lock.
Argument BOUND is a buffer position limiting searching."
    (let ( ;=====================================================================
		point
		(case-fold-search nil)
	) ;==========================================================================
        (while 
			(and 
			    (not point)
				(re-search-forward "\\(::[ \t\n]*{\\|&\\)[ \t]*" bound t)
			)
		    (while 
				(or 
			        (re-search-forward "\\=\n[ \t]*" bound t)
					(re-search-forward "\\=%.*" bound t)
					(and 
				        (re-search-forward "\\=/\\*" bound t)
						(re-search-forward "\\*/[ \t]*" bound t))
				)
		    )
			(setq point 
			    (re-search-forward
				    (format "\\=\\(%s\\)" mercury-atom-regexp) bound t))
		)
		point
	) ;==========================================================================
)

;;-------------------------------------------------------------------
;; Indentation stuff
;;-------------------------------------------------------------------

;; NB: This function *MUST* have this optional argument since XEmacs
;; assumes it. This does not mean we have to use it...
(defun mercury-indent-line (&optional whole-exp)
    "Indent current line as Mercury code.
With argument, indent any additional lines of the same clause
rigidly along with this one (not yet)."
	(interactive "p")
	(let ( ;=====================================================================
		(indent (mercury-indent-level))
        (pos (- (point-max) (point))) 
		beg
	) ;==========================================================================
	    (beginning-of-line)
		(setq beg (point))
		(skip-chars-forward " \t")
		(if (zerop (- indent (current-column)))
			nil
		    (delete-region beg (point))
			(indent-to indent))
		(if (> (- (point-max) pos) (point))
			(goto-char (- (point-max) pos)))
		;; Align comments
		(if mercury-align-comments-flag
			(save-excursion (mercury-goto-comment-column t)))
		;; Insert spaces if needed
		(if mercury-electric-tab-flag
			(mercury-insert-spaces-after-paren))
    ) ;==========================================================================
)

(defun mercury-comment-indent ()
    "Compute mercury comment indentation."
    (cond 
	    ((looking-at "%%%") 
		    (mercury-indentation-level-of-line))
        ((looking-at "%%") 
		    (mercury-indent-level))
        (t
            (save-excursion
			    (skip-chars-backward " \t")
				;; Insert one space at least, except at left margin.
				(max 
				    (+ (current-column) (if (bolp) 0 1))
					comment-column
				))))
)

(defun mercury-indent-level ()
    "Compute mercury indentation level."
	(save-excursion
        (beginning-of-line)
		(let ( ;=================================================================
			(totbal 
			    (mercury-region-paren-balance (mercury-clause-start t) (point)))
			(oldpoint (point))
		) ;======================================================================
		    (skip-chars-forward " \t")
			(cond
			    ((looking-at "%%%") 
				    (mercury-indentation-level-of-line))
                                        ;Large comment starts
				((looking-at "%[^%]") 
				    comment-column) ;Small comment starts
				((bobp) 
				    0)                        ;Beginning of buffer
				;; If we found '}' then we must check if it's the
				;; end of an object declaration or something else.
				((and 
				    (looking-at "}") 
					(save-excursion
					    (forward-char 1)
						;; Goto to matching {
						(if mercury-use-mercury-tokenizer-flag
							(mercury-backward-list)
						    (backward-list))
						(skip-chars-backward " \t")
						(backward-char 2)
						(looking-at "::"))
				)
				    ;; It was an object
				    (if mercury-object-end-to-0-flag
						0
					    mercury-indent-width))
				;;End of /* */ comment
				((looking-at "\\*/")                  
				    (save-excursion
					    (mercury-find-start-of-mline-comment)
						(skip-chars-backward " \t")
						(- (current-column) 2)
					))
				;; Here we check if the current line is within a /* */ pair
				((and 
				    (looking-at "[^%/]")
					(eq (mercury-in-string-or-comment) 'cmt)
				) 
				    (if mercury-indent-mline-comments-flag
						(mercury-find-start-of-mline-comment)
					    ;; Same as before
					    (mercury-indentation-level-of-line)))

				(t
				    (let ( ;=====================================================
						(empty t) 
						ind 
						linebal
					) ;==========================================================
					    ;; See previous indentation
					    (while empty
						    (forward-line -1)
							(beginning-of-line)
							(if (= (point) (point-min))
								(setq empty nil)
							    (skip-chars-forward " \t")
								(if 
									(not (or 
										(not (member (mercury-in-string-or-comment) '(nil txt)))
										(looking-at "%") 
										(looking-at "\n")
									))
									(setq empty nil)))
						)
						;; Store this line's indentation
						(if (= (point) (point-min))
							(setq ind 0)                ;Beginning of buffer
						    (setq ind (current-column))) ;Beginning of clause
						;; Compute the balance of the line
						(setq linebal (mercury-paren-balance))
						;;(message "bal of previous line %d totbal %d" linebal totbal)
						(if (< linebal 0)
							(progn
							    ;; Add 'indent-level' mode to find-unmatched-paren instead?
							    (end-of-line)
								(setq ind (mercury-find-indent-of-matching-paren))
							))
						;;(message "ind %d" ind)
						(beginning-of-line)
;; <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
						;; Check if the line ends with ":-", ".", ":: {", "}" (might be
						;; unnecessary), "&" or ")" (The last four concerns SICStus objects)
						(cond
						    ;; Increase indentation if the previous line was the head of a rule
						    ;; and does not contain a '.'
						    ((and 
							    (looking-at 
								    (format ".*%s[^\\.]*[ \t]*\\(%%.*\\|\\)$" 
										mercury-head-delimiter))
								;; We must check that the match is at a paren balance of 0.
								(save-excursion
								    (let (
										(p (point))
									)
									    (re-search-forward mercury-head-delimiter)
										(>= 0 (mercury-region-paren-balance p (point)))
									)
								)
							)
							    (let (
									headindent
								)
								    (if (< (mercury-paren-balance) 0)
										(save-excursion
										    (end-of-line)
											(setq headindent (mercury-find-indent-of-matching-paren))
										)
									    (setq headindent (mercury-indentation-level-of-line)))
									    (setq ind (+ headindent mercury-indent-width))
								))
							;; The previous line was the head of an object
							((looking-at ".+ *::.*{[ \t]*$")
							    (setq ind mercury-indent-width))
							;; If a '.' is found at the end of the previous line, then
							;; decrease the indentation. (The \\(%.*\\|\\) part of the
							;; regexp is for comments at the end of the line)
							((and 
							    (looking-at "^.+\\.[ \t]*\\(%.*\\|\\)$") 
								;; Make sure that the '.' found is not in a comment or string
								(save-excursion
								    (end-of-line)
									(re-search-backward "\\.[ \t]*\\(%.*\\|\\)$" (point-min))
									;; Guard against the real '.' being followed by a
									;; commented '.'.
									(if (eq (mercury-in-string-or-comment) 'cmt)  ;; commented out '.'
										(let (
											(here 
											    (save-excursion
												    (beginning-of-line)
													(point)
												))
										)
										    (end-of-line)
											(re-search-backward "\\.[ \t]*%.*$" here t)
										)
									    (not (mercury-in-string-or-comment)))
							    )
							)
							    (setq ind 0))
							
							;; If a '.' is found at the end of the previous line, then
							;; decrease the indentation. (The /\\*.*\\*/ part of the
							;; regexp is for C-like comments at the end of the
							;; line--can we merge with the case above?).
							((and 
							    (looking-at "^.+\\.[ \t]*\\(/\\*.*\\|\\)$") 
								;; Make sure that the '.' found is not in a comment or string
								(save-excursion
								    (end-of-line)
									(re-search-backward "\\.[ \t]*\\(/\\*.*\\|\\)$" (point-min))
									;; Guard against the real '.' being followed by a
									;; commented '.'.
									(if (eq (mercury-in-string-or-comment) 'cmt)  ;; commented out '.'
										(let (
											(here 
											    (save-excursion
												    (beginning-of-line)
													(point)
												))
										)
										    (end-of-line)
											(re-search-backward "\\.[ \t]*/\\*.*$" here t)
										)
									    (not (mercury-in-string-or-comment)))
								)
							)
							    (setq ind 0))

						)
						;; If the last non comment char is a ',' or left paren or a left-
						;; indent-regexp then indent to open parenthesis level
						(if (> totbal 0)
						    (if 
								(looking-at
								    (format 
									    "\\(%s\\|%s\\|0'.\\|[0-9]+'[0-9a-zA-Z]+\\|[^\n\'\"%%]\\)*\\(,\\|%s\\|%s\\)\[ \t]*\\(%%.*\\|\\)$" 
										mercury-quoted-atom-regexp mercury-string-regexp
										mercury-left-paren mercury-left-indent-regexp))
								(progn
								    (goto-char oldpoint)
									(setq ind 
										(mercury-find-unmatched-paren 
										    (if mercury-paren-indent-p 
												'termdependent 
											    'skipwhite)))
									;;(setq ind (mercury-find-unmatched-paren 'termdependent))
								)
							  (goto-char oldpoint)
							  (setq ind (mercury-find-unmatched-paren nil))))
						;; Return the indentation level
						ind
                    )
				)
			)
		)
	)
)

(defun mercury-find-indent-of-matching-paren ()
    "Find the indentation level based on the matching parenthesis.
Indentation level is set to the one the point is after when the function is
called."
    (save-excursion
        ;; Go to the matching paren
        (if mercury-use-mercury-tokenizer-flag
            (mercury-backward-list)
		    (backward-list))
		;; If this was the first paren on the line then return this line's
		;; indentation level
		(if (mercury-paren-is-the-first-on-line-p)
			(mercury-indentation-level-of-line)
		    ;; It was not the first one
		    (progn
			    ;; Find the next paren
			    (mercury-goto-next-paren 0)
				;; If this paren is a left one then use its column as indent level,
				;; if not then recurse this function
				(if (looking-at mercury-left-paren)
					(+ (current-column) 1)
				    (progn
					    (forward-char 1)
						(mercury-find-indent-of-matching-paren)
					))
			))
    )
)

(defun mercury-indentation-level-of-line ()
    "Return the indentation level of the current line."
    (save-excursion
	    (beginning-of-line)
		(skip-chars-forward " \t")
		(current-column)
	)
)

(defun mercury-first-pos-on-line ()
    "Return the first position on the current line."
	(save-excursion
	    (beginning-of-line)
		(point)
	)
)

(defun mercury-paren-is-the-first-on-line-p ()
    "Return t if the parenthesis under the point is the first one on the line.
Return nil otherwise.
Note: does not check if the point is actually at a parenthesis!"
    (save-excursion
        (let (
			(begofline (mercury-first-pos-on-line))
		)
		    (if (= begofline (point))
				t
			    (if (mercury-goto-next-paren begofline)
					nil
				    t))
		)
	)
)

(defun mercury-find-unmatched-paren (&optional mode)
    "Return the column of the last unmatched left parenthesis.
If MODE is `skipwhite' then any white space after the parenthesis is added to
the answer.
If MODE is `plusone' then the parenthesis' column +1 is returned.
If MODE is `termdependent' then if the unmatched parenthesis is part of
a compound term the function will work as `skipwhite', otherwise
it will return the column paren plus the value of `mercury-paren-indent'.
If MODE is nil or not set then the parenthesis' exact column is returned."
    (save-excursion
	    ;; If the next paren we find is a left one we're finished, if it's
	    ;; a right one then we go back one step and recurse
	    (mercury-goto-next-paren 0)
		(let (
			(roundparen (looking-at "("))
		)
		    (if (looking-at mercury-left-paren)
				(let (
					(not-part-of-term 
					    (save-excursion (backward-char 1) (looking-at "[ \t]")))
				)
				    (if (eq mode nil)
						(current-column)
					    (if (and roundparen (eq mode 'termdependent) not-part-of-term)
							(+ (current-column)
							   (if mercury-electric-tab-flag
								   ;; Electric TAB
								   mercury-paren-indent
								   ;; Not electric TAB
								   (if (looking-at ".[ \t]*$")
									   2
									   mercury-paren-indent)))
						    (forward-char 1)
							(if (or (eq mode 'skipwhite) (eq mode 'termdependent) )
								(skip-chars-forward " \t")
							)
							(current-column)))
				)
			    ;; Not looking at left paren
			    (progn
				    (forward-char 1)
					;; Go to the matching paren. When we get there we have a total
					;; balance of 0.
					(if mercury-use-mercury-tokenizer-flag
						(mercury-backward-list)
					    (backward-list))
					(mercury-find-unmatched-paren mode)
				))
		)
	)
)


(defun mercury-paren-balance ()
    "Return the parenthesis balance of the current line.
A return value of n means n more left parentheses than right ones."
    (save-excursion
	    (end-of-line)
		(mercury-region-paren-balance (mercury-first-pos-on-line) (point))
	)
)

(defun mercury-region-paren-balance (beg end)
    "Return the summed parenthesis balance in the region.
The region is limited by BEG and END positions."
    (save-excursion
        (let (
			(state 
			    (if mercury-use-mercury-tokenizer-flag
					(mercury-tokenize beg end)
				    (parse-partial-sexp beg end)))
		)
		    (nth 0 state)
		)
	)
)

(defun mercury-goto-next-paren (limit-pos)
    "Move the point to the next parenthesis earlier in the buffer.
Return t if a match was found before LIMIT-POS.  Return nil otherwise."
    (let (
		retval
	)
	    (setq retval 
			(re-search-backward
			    (concat mercury-left-paren "\\|" mercury-right-paren) limit-pos t))
		;; If a match was found but it was in a string or comment, then recurse
		(if (and retval (mercury-in-string-or-comment))
			(mercury-goto-next-paren limit-pos)
		    retval)
    )
)

(defun mercury-in-string-or-comment ()
    "Check whether string, atom, or comment is under current point.
Return:
 `txt' if the point is in a string, atom, or character code expression
 `cmt' if the point is in a comment
 nil otherwise."
    (save-excursion
        (let* (
			(start
			    (if (eq mercury-parse-mode 'beg-of-line)
					;; 'beg-of-line
					(save-excursion
					    (let (
							safepoint
						)
						    (beginning-of-line)
							(setq safepoint (point))
							(while 
								(and 
								    (> (point) (point-min))
									(progn
                                        (forward-line -1)
										(end-of-line)
										(if (not (bobp)) 
											(backward-char 1))
										(looking-at "\\\\")
									))
							    (beginning-of-line)
								(setq safepoint (point)))
							safepoint
						)
					)
				    ;; 'beg-of-clause
				    (mercury-clause-start)))
			(end (point))
			(state 
			    (if mercury-use-mercury-tokenizer-flag
					(mercury-tokenize start end)
				    (parse-partial-sexp start end)))
		)
		    (cond
			    ((nth 3 state) 'txt) ; String
				((nth 4 state) 'cmt) ; Comment
				(t
				    (cond
					    ((looking-at "%") 'cmt) ; Start of a comment
						((looking-at "/\\*") 'cmt) ; Start of a comment
						((looking-at "\'") 'txt) ; Start of an atom
						((looking-at "\"") 'txt) ; Start of a string
						(t nil))))
		)
    )
)

(defun mercury-find-start-of-mline-comment ()
    "Return the start column of a /* */ comment.
This assumes that the point is inside a comment."
    (re-search-backward "/\\*" (point-min) t)
    (forward-char 2)
    (skip-chars-forward " \t")
    (current-column)
)

(defun mercury-insert-spaces-after-paren ()
    "Insert spaces after the opening parenthesis.
Spaces are inserted if the parenthesis is the first non-white character
on the line."
    (save-excursion
	    (let (
		    incr 
		    level
	    )
		    (beginning-of-line)
			(skip-chars-forward " \t")
			(if (looking-at (format "(\\|%s" mercury-left-indent-regexp))
			    (progn
				    ;; Treat "( If -> " lines specially.
				    (if (looking-at "(.*->")
					    (setq incr 2)
					    (setq incr mercury-paren-indent))
					;; Goto the first char after the matched text
					(re-search-forward 
				        (format "(\\|%s" mercury-left-indent-regexp)
					    (point-max) 
					    t)
					(goto-char (match-end 0))
					(setq level (+ (mercury-find-unmatched-paren) incr))
					;; Remove old white space
					(let (
						(start (point))
					)
					    (skip-chars-forward " \t")
						(delete-region start (point))
					)
					(indent-to level)
				))
		)
	)
	(skip-chars-forward " \t")
)


;;-------------------------------------------------------------------
;; The tokenizer
;;-------------------------------------------------------------------


(defun mercury-tokenize (beg end &optional stopcond)
    "Tokenize a region of mercury code between BEG and END.
STOPCOND decides the stop condition of the parsing. Valid values
are 'zerodepth which stops the parsing at the first right parenthesis
where the parenthesis depth is zero, 'skipover which skips over
the current entity (e.g. a list, a string, etc.) and nil.

The function returns a list with the following information:
 0. parenthesis depth 
 3. 'atm if END is inside an atom
    'str if END is inside a string
    'chr if END is in a character code expression (0'x)
    nil otherwise
 4. non-nil if END is inside a comment
 5. end position (always equal to END if STOPCOND is nil)
The rest of the elements are undefined."
    (save-excursion
	    (let* ( ;================================================================
			(end2 (1+ end))
			oldp
			(depth 0)
			(quoted nil)
			inside_cmt
			(endpos end2)
			skiptype ; The type of entity we'll skip over
        ) ;======================================================================
		    (goto-char beg)
			(if (and (eq stopcond 'skipover) (looking-at "[^[({'\"]"))
			    (setq endpos (point))                ; Stay where we are
			    (while 
					(and
				        (re-search-forward mercury-tokenize-searchkey end2 t)
						(< (point) end2))
				    (progn
					    (setq oldp (point))
						(goto-char (match-beginning 0))
						(cond
					        ;; Atoms and strings
						    ((looking-at "'")
							    ;; Find end of atom
							    (if (re-search-forward "[^\\]'" end2 'limit)
									;; Found end of atom
									(progn
									    (setq oldp end2)
										(if (and (eq stopcond 'skipover) (not skiptype))
											(setq endpos (point))
										    (setq oldp (point)))
									) ; Continue tokenizing
								    (setq quoted 'atm)))

							((looking-at "\"")
							    ;; Find end of string
							    (if (re-search-forward "[^\\]\"" end2 'limit)
									;; Found end of string
									(progn
									    (setq oldp end2)
										(if (and (eq stopcond 'skipover) (not skiptype))
											(setq endpos (point))
										    (setq oldp (point)))
									) ; Continue tokenizing
								    (setq quoted 'str)))

							;; Paren stuff
							((looking-at mercury-left-paren)
							    (setq depth (1+ depth))
								(setq skiptype 'paren))

							((looking-at mercury-right-paren)
							    (setq depth (1- depth))
								(if 
									(and
									    (or 
										    (eq stopcond 'zerodepth)
											(and (eq stopcond 'skipover) (eq skiptype 'paren)))
										(= depth 0))
									(progn
									    (setq endpos (1+ (point)))
										(setq oldp end2)
									)))

							;; Comment stuff
							((looking-at comment-start)
							    (end-of-line)
								;; (if (>= (point) end2)
								(if (>= (point) end)
									(progn
									    (setq inside_cmt t)
										(setq oldp end2)
									)
								    (setq oldp (point))))

							((looking-at "/\\*")
							    (if (re-search-forward "\\*/" end2 'limit)
									(setq oldp (point))
								    (setq inside_cmt t)
									(setq oldp end2)))

							;; 0'char
							((looking-at "0'")
							    (setq oldp (1+ (match-end 0)))
								(if (> oldp end) 
									(setq quoted 'chr)))
           
							;; base'number
							((looking-at "[0-9]+'")
							    (goto-char (match-end 0))
								(skip-chars-forward "0-9a-zA-Z")
								(setq oldp (point)))

             
						) ; cond
						(goto-char oldp)
					) ; progn
				)) ; End of while, end of if
			;; Create return list
			(list depth nil nil quoted inside_cmt endpos)
        ) ;======================================================================
	)
)

;;-------------------------------------------------------------------
;; Miscellaneous functions
;;-------------------------------------------------------------------

;; For Windows. Change backslash to slash. SICStus handles either
;; path separator but backslash must be doubled, therefore use slash.
(defun mercury-bsts (string)
    "Change backslashes to slashes in STRING."
	(let (
		(str1 (copy-sequence string))
        (len (length string))
        (i 0)
	)
        (while (< i len)
		    (if (char-equal (aref str1 i) ?\\)
				(aset str1 i ?/))
			(setq i (1+ i)))
		str1
	)
)

;(defun mercury-temporary-file ()
;  "Make temporary file name for compilation."
;  (make-temp-name 
;   (concat 
;    (or
;     (getenv "TMPDIR")
;     (getenv "TEMP") 
;     (getenv "TMP")
;     (getenv "SYSTEMP")
;     "/tmp")
;    "/prolcomp")))
;(setq mercury-temp-filename (mercury-bsts (mercury-temporary-file)))

(defun mercury-temporary-file ()
    "Make temporary file name for compilation."
    (let* (
		(umask  (default-file-modes))
        (temporary-file-directory 
		    (or
			    (getenv "TMPDIR")
				(getenv "TEMP") 
				(getenv "TMP")
				(getenv "SYSTEMP")
				"/tmp"))
        (prefix (expand-file-name "prolcomp" temporary-file-directory))
        (suffix ".pl")
        file
	)
	    (unwind-protect
			(progn
			    ;; Create temp files with strict access rights.
			    (set-default-file-modes #o700)
				(while 
					(condition-case ()
                        (progn
						    (setq file (concat (make-temp-name prefix) suffix))
							(unless (file-exists-p file) 
							    (write-region "" nil file nil 'silent))
							nil
						)
					    (file-already-exists t))
				    ;; the file was somehow created by someone else between
				    ;; `make-temp-name' and `write-region', let's try again.
				    nil)
				file
			)
		    ;; Reset the umask.
		    (set-default-file-modes umask)
		)
	)
)

(defun mercury-goto-mercury-process-buffer ()
    "Switch to the mercury process buffer and go to its end."
    (switch-to-buffer-other-window "*mercury*")
    (goto-char (point-max))
)

(defun mercury-get-predspec ()
    (save-excursion
	    (let (
			(state (mercury-clause-info))
			(object (mercury-in-object))
		)
		    (if (equal (nth 0 state) "")
				nil
			    (format "%s/%d" (nth 0 state) (nth 1 state)))
		)
	)
)

;; For backward compatibility. Stolen from custom.el.
(or 
    (fboundp 'match-string)
    ;; Introduced in Emacs 19.29.
    (defun match-string (num &optional string)
        "Return string of text matched by last search.
NUM specifies which parenthesized expression in the last regexp.
 Value is nil if NUMth pair didn't match, or there were less than NUM pairs.
Zero means the entire text matched by the whole regexp or whole string.
STRING should be given if the last search was by `string-match' on STRING."
		(if (match-beginning num)
			(if string
				(substring string (match-beginning num) (match-end num))
			    (buffer-substring (match-beginning num) (match-end num))))
	)
)

(defun mercury-pred-start ()
    "Return the starting point of the first clause of the current predicate."
    (save-excursion
	    (goto-char (mercury-clause-start))
		;; Find first clause, unless it was a directive
		(if (and (not (looking-at "[:?]-")) (not (looking-at "[ \t]*[%/]"))) ; Comment
			(let* (
				(pinfo (mercury-clause-info))
				(predname (nth 0 pinfo))
				(arity (nth 1 pinfo))
				(op (point))
			)
			    (while 
					(and 
					    (re-search-backward
						    (format "^%s\\([(\\.]\\| *%s\\)" predname mercury-head-delimiter) nil t)
						(= arity (nth 1 (mercury-clause-info))))
				    (setq op (point)))
				;; Skip to the beginning of declarations of the predicate
				(progn
				    (goto-char (mercury-beginning-of-clause))
					(while 
						(and 
						    (not (eq (point) op))
                            (looking-at
                                (format ":-[ \t]*\\(pred\\|mode\\)[ \t]+%s" predname)))
					    (setq op (point))
						(goto-char (mercury-beginning-of-clause)))
				)
				op
			)
		    (point))
    )
)

(defun mercury-pred-end ()
    "Return the position at the end of the last clause of the current predicate."
	(save-excursion
	    (goto-char (mercury-clause-end))        ; if we are before the first predicate
		(goto-char (mercury-clause-start))
		(let* (
			(pinfo (mercury-clause-info))
			(predname (nth 0 pinfo))
			(arity (nth 1 pinfo))
			oldp
			(notdone t)
			(op (point))
		)
		    (if (looking-at "[:?]-")
				;; This was a directive
				(progn
				    (if (looking-at (format ":-[ \t]*\\(pred\\|mode\\)[ \t]+\\(%s+\\)" mercury-atom-regexp))
						;; Skip predicate declarations
						(progn
						    (setq predname 
								(buffer-substring-no-properties (match-beginning 2) (match-end 2)))
							(while 
								(re-search-forward
								    (format "\n*\\(:-[ \t]*\\(pred\\|mode\\)[ \t]+\\)?%s[( \t]" predname)
									nil 
									t))
						))
					(goto-char (mercury-clause-end))
					(setq op (point))
				)
			  ;; It was not a directive, find the last clause
			  (while 
				  (and 
				      notdone
					  (re-search-forward
					      (format "^%s\\([(\\.]\\| *%s\\)" predname mercury-head-delimiter) nil t)
					  (= arity (nth 1 (mercury-clause-info))))
				(setq oldp (point))
				(setq op (mercury-clause-end))
				(if (>= oldp op)
					;; End of clause not found.
					(setq notdone nil)
				    ;; Continue while loop
				    (goto-char op))))
			op
		)
	)
)

(defun mercury-clause-start (&optional not-allow-methods)
    "Return the position at the start of the head of the current clause."
	(save-excursion
        (let (
			(notdone t)
			(retval (point-min))
		)
		    (end-of-line)
			(while 
				(and 
				    notdone 
					;; Search for a text at beginning of a line
					;; ######
					;; (re-search-backward "^[a-z$']" nil t))
					(re-search-backward
					    ;; (format "^[%s$']" mercury-lower-case-string)
					    (format "^\\([%s$']\\|[:?]-\\)" mercury-lower-case-string) nil t))
			    (let (
					(bal (mercury-paren-balance))
				)
				    (cond
					    ((> bal 0)
						    ;; Start of clause found
						    (progn
							    (setq retval (point))
								(setq notdone nil)))

						((and 
						    (= bal 0)
							(looking-at
							    (format ".*\\(\\.\\|%s\\|!,\\)[ \t]*\\(%%.*\\|\\)$" mercury-head-delimiter))
						)
						    ;; Start of clause found if the line ends with a '.' or
						    ;; a mercury-head-delimiter
						    (progn
							    (setq retval (point))
								(setq notdone nil)))

						(t nil)) ; Do nothing
				)
			)    
			retval
		)
	)
)

(defun mercury-clause-end (&optional not-allow-methods)
    "Return the position at the end of the current clause."
	(save-excursion
	    (beginning-of-line)                ; Necessary since we use "^...." for the search
		(if 
			(re-search-forward 
			    (format
				    "^\\(%s\\|%s\\|[^\n\'\"%%]\\)*\\.[ \t]*\\(\\|%%.*\\)$"
					mercury-quoted-atom-regexp mercury-string-regexp)
				nil 
				t)
			(if (and (mercury-in-string-or-comment) (not (eobp)))
				(progn
				    (forward-char)
					(mercury-clause-end)
				)
			    (point))
		    (point))
	)
)

(defun mercury-clause-info ()
    "Return a (name arity) list for the current clause."
	(let (
		predname 
		(arity 0)
	)
	    (save-excursion
		    (goto-char (mercury-clause-start))
			(let (
				(op (point))
			)
			    (if (looking-at mercury-atom-char-regexp)
					(progn
					    (skip-chars-forward "^ (\\.")
						(setq predname (buffer-substring op (point)))
					)
				    (setq predname ""))
				;; Retrieve the arity
				(if (looking-at mercury-left-paren)
					(let (
						(endp (save-excursion (mercury-forward-list) (point)))
					)
					    (setq arity 1)
						(forward-char 1)                ; Skip the opening paren
						(while 
							(progn 
							    (skip-chars-forward "^[({,'\"") 
								(< (point) endp))
						    (if (looking-at ",")
								(progn
								    (setq arity (1+ arity))
									(forward-char 1))        ; Skip the comma
							    ;; We found a string, list or something else we want
							    ;; to skip over. Always use mercury-tokenize,
							    ;; parse-partial-sexp does not have a 'skipover mode.
							    (goto-char (nth 5 (mercury-tokenize (point) endp 'skipover))))
						)
					)
				)
				(list predname arity)
			)
		)
	)
)

(defun mercury-in-object ()
    "Return object name if the point is inside a SICStus object definition."
	;; Return object name if the last line that starts with a character
	;; that is neither white space nor a comment start
	(save-excursion
	    (if 
			(save-excursion 
			    (beginning-of-line)
				(looking-at "\\([^\n ]+\\)[ \t]*::[ \t]*{"))
			;; We were in the head of the object
			(match-string 1)
		    ;; We were not in the head
		    (if (and (re-search-backward "^[a-z$'}]" nil t) (looking-at "\\([^\n ]+\\)[ \t]*::[ \t]*{"))
				(match-string 1)
			    nil))
	)
)

(defun mercury-forward-list ()
    "Move the point to the matching right parenthesis."
    (interactive)
   (if mercury-use-mercury-tokenizer-flag
	   (let (
		   (state (mercury-tokenize (point) (point-max) 'zerodepth))
	   )
		   (goto-char (nth 5 state))
	   )
	   (forward-list))
)

;; NB: This could be done more efficiently!
(defun mercury-backward-list ()
    "Move the point to the matching left parenthesis."
    (interactive)
    (if mercury-use-mercury-tokenizer-flag
		(let (
			(bal 0)
            (paren-regexp (concat mercury-left-paren "\\|" mercury-right-paren))
            (notdone t)
		)
		    (while (and notdone (re-search-backward paren-regexp nil t))
			    (cond
				    ((looking-at mercury-left-paren)
					    (if (not (mercury-in-string-or-comment))
							(setq bal (1+ bal)))
						(if (= bal 0)
							(setq notdone nil)))
					((looking-at mercury-right-paren)
					    (if (not (mercury-in-string-or-comment))
							(setq bal (1- bal))))))
		)
	    (backward-list))
)

(defun mercury-beginning-of-clause ()
    "Move to the beginning of current clause.
If already at the beginning of clause, move to previous clause."
	(interactive)
	(let (
		(point (point))
        (new-point (mercury-clause-start))
	)
	    (if (and (>= new-point point) (> point 1))
			(progn
			    (goto-char (1- point))
				(goto-char (mercury-clause-start))
			)
		    (goto-char new-point)
			(skip-chars-forward " \t"))
	)
)

;; (defun mercury-previous-clause ()
;;   "Move to the beginning of the previous clause."
;;   (interactive)
;;   (forward-char -1)
;;   (mercury-beginning-of-clause))

(defun mercury-end-of-clause ()
    "Move to the end of clause.
If already at the end of clause, move to next clause."
	(interactive)
	(let (
		(point (point))
		(new-point (mercury-clause-end))
	)
	    (if (and (<= new-point point) (not (eq new-point (point-max))))
			(progn
			    (goto-char (1+ point))
				(goto-char (mercury-clause-end))
			)
		    (goto-char new-point)
		)
	)
)

;; (defun mercury-next-clause ()
;;   "Move to the beginning of the next clause."
;;   (interactive)
;;   (mercury-end-of-clause)
;;   (forward-char)
;;   (mercury-end-of-clause)
;;   (mercury-beginning-of-clause))

(defun mercury-beginning-of-predicate ()
    "Go to the nearest beginning of predicate before current point.
Return the final point or nil if no such a beginning was found."
	(interactive)
	(let (
		(op (point))
        (pos (mercury-pred-start))
	)
	    (if pos
			(if (= op pos)
				(if (not (bobp))
					(progn
					    (goto-char pos)
						(backward-char 1)
						(setq pos (mercury-pred-start))
						(if pos
							(progn
							    (goto-char pos)
								(point)))
					))
			    (goto-char pos)
				(point)))
	)
)

(defun mercury-end-of-predicate ()
    "Go to the end of the current predicate."
	(interactive)
	(let (
		(op (point))
	)
	    (goto-char (mercury-pred-end))
		(if (= op (point))
			(progn
			    (forward-line 1)
				(mercury-end-of-predicate)
			))
	)
)

(defun mercury-insert-predspec ()
    "Insert the predspec for the current predicate."
    (interactive)
    (let* (
		(pinfo (mercury-clause-info))
        (predname (nth 0 pinfo))
        (arity (nth 1 pinfo))
	)
	    (insert (format "%s/%d" predname arity))
	)
)

(defun mercury-view-predspec ()
    "Insert the predspec for the current predicate."
	(interactive)
	(let* (
		(pinfo (mercury-clause-info))
		(predname (nth 0 pinfo))
		(arity (nth 1 pinfo))
	)
	    (message (format "%s/%d" predname arity))
	)
)

(defun mercury-insert-predicate-template ()
    "Insert the template for the current clause."
	(interactive)
	(let* (
		(n 1)
		oldp
		(pinfo (mercury-clause-info))
		(predname (nth 0 pinfo))
		(arity (nth 1 pinfo))
	)
	    (insert predname)
		(if (> arity 0)
			(progn
			    (insert "(")
				(setq oldp (point))
				(while (< n arity)
				    (insert ",")
					(setq n (1+ n)))
				(insert ")")
				(goto-char oldp)
			))
	)
)

(defun mercury-insert-next-clause ()
    "Insert newline and the name of the current clause."
	(interactive)
	(insert "\n")
	(mercury-insert-predicate-template)
)

(defun mercury-insert-module-modeline ()
    "Insert a modeline for module specification.
This line should be first in the buffer.
The module name should be written manually just before the semi-colon."
	(interactive)
	(insert "%%% -*- Module: ; -*-\n")
	(backward-char 6)
)

(defun mercury-uncomment-region (beg end)
    "Uncomment the region between BEG and END."
	(interactive "r")
	(comment-region beg end -1)
)

(defun mercury-goto-comment-column (&optional nocreate)
    "Move comments on the current line to the correct position.
If NOCREATE is nil (or omitted) and there is no comment on the line, then
a new comment is created."
	(interactive)
	(beginning-of-line)
	(if 
		(or 
		    (not nocreate)
			(and
			    (re-search-forward 
				    (format "^\\(\\(%s\\|%s\\|[^\n\'\"%%]\\)*\\)%% *" 
						mercury-quoted-atom-regexp mercury-string-regexp)
					(save-excursion (end-of-line) (point)) 
					'limit)
				(progn
				    (goto-char (match-beginning 0))
					(not (eq (mercury-in-string-or-comment) 'txt)))))
		(indent-for-comment))
)

(defun mercury-indent-predicate ()
    "*Indent the current predicate."
	(interactive)
	(indent-region (mercury-pred-start) (mercury-pred-end) nil)
)

(defun mercury-indent-buffer ()
    "*Indent the entire buffer."
    (interactive)
    (indent-region (point-min) (point-max) nil)
)

(defun mercury-mark-clause ()
    "Put mark at the end of this clause and move point to the beginning."
	(interactive)
	(let (
		(pos (point))
	)
	    (goto-char (mercury-clause-end))
		(forward-line 1)
		(beginning-of-line)
		(set-mark (point))
		(goto-char pos)
		(goto-char (mercury-clause-start))
	)
)

(defun mercury-mark-predicate ()
    "Put mark at the end of this predicate and move point to the beginning."
	(interactive)
	(let (
		pos
	)
	    (goto-char (mercury-pred-end))
		(setq pos (point))
		(forward-line 1)
		(beginning-of-line)
		(set-mark (point))
		(goto-char pos)
		(goto-char (mercury-pred-start))
	)
)

;; Stolen from `cc-mode.el':
(defun mercury-electric-delete (arg)
    "Delete preceding character or whitespace.
If `mercury-hungry-delete-key-flag' is non-nil, then all preceding whitespace is
consumed.  If however an ARG is supplied, or `mercury-hungry-delete-key-flag' is
nil, or point is inside a literal then the function in the variable
`backward-delete-char' is called."
	(interactive "P")
	(if 
		(or 
		    (not mercury-hungry-delete-key-flag)
			arg
			(mercury-in-string-or-comment))
		(funcall 'backward-delete-char (prefix-numeric-value arg))
	    (let (
			(here (point))
		)
		    (skip-chars-backward " \t\n")
			(if (/= (point) here)
				(delete-region (point) here)
			    (funcall 'backward-delete-char 1))
		))
)

;; For XEmacs compatibility (suggested by Per Mildner)
(put 'mercury-electric-delete 'pending-delete 'supersede)

(defun mercury-electric-dot (arg)
    "Insert dot and newline or a head of a new clause.

If `mercury-electric-dot-flag' is nil, then simply insert dot.
Otherwise::
When invoked at the end of nonempty line, insert dot and newline.
When invoked at the end of an empty line, insert a recursive call to
the current predicate.
When invoked at the beginning of line, insert a head of a new clause
of the current predicate.

When called with prefix argument ARG, insert just dot."
	(interactive "P")
	;; Check for situations when the electricity should not be active
	(if 
		(or 
		    (not mercury-electric-dot-flag)
			arg
			(mercury-in-string-or-comment)
			;; Do not be electrix in a floating point number or an operator
			(save-excursion 
			    (not 
				    (re-search-backward
					    ;; ######
					    ;; "\\(^\\|[])}a-zA-Z_!'0-9]+\\)[ \t]*\\=" nil t)))
					    (format 
						    "\\(^\\|[])}%s%s_!'0-9]+\\)[ \t]*\\=" 
							mercury-lower-case-string
							mercury-upper-case-string)
						nil 
						t))
			)
			;; Do not be electric if inside a parenthesis pair.
			(not (= (mercury-region-paren-balance (mercury-clause-start) (point)) 0)))
		(funcall 'self-insert-command (prefix-numeric-value arg))
	    (cond
		    ;; Beginning of line
		    ((bolp)
			    (mercury-insert-predicate-template))
			;; At an empty line with at least one whitespace
			((save-excursion (beginning-of-line) (looking-at "[ \t]+$"))
			    (mercury-insert-predicate-template)
				(save-excursion (end-of-line) (insert ".\n")))
			;; Default
			(t
			    (insert ".\n"))))
)

(defun mercury-electric-underscore ()
    "Replace variable with an underscore.
If `mercury-electric-underscore-flag' is non-nil and the point is
on a variable then replace the variable with underscore and skip
the following comma and whitespace, if any.
If the point is not on a variable then insert underscore."
	(interactive)
	(if mercury-electric-underscore-flag
		(let (;start
            (oldcase case-fold-search)
            (oldp (point))
		)
		    (setq case-fold-search nil)
			;; ######
			;;(skip-chars-backward "a-zA-Z_")
			(skip-chars-backward
			    (format "%s%s_" mercury-lower-case-string mercury-upper-case-string))
			;(setq start (point))
			(if 
				(and 
				    (not (mercury-in-string-or-comment))
					;; ######
					;; (looking-at "\\<[_A-Z][a-zA-Z_0-9]*\\>"))
					(looking-at 
					    (format 
						    "\\<[_%s][%s%s_0-9]*\\>"  
							mercury-upper-case-string
							mercury-lower-case-string
							mercury-upper-case-string)))
				(progn
				    (replace-match "_")
					(skip-chars-forward ", \t\n")
				)
			    (goto-char oldp)
				(self-insert-command 1))
			(setq case-fold-search oldcase)
		)
	    (self-insert-command 1))
  )


(defun mercury-find-term (functor arity &optional prefix)
    "Go to the position at the start of the next occurance of a term.
The term is specified with FUNCTOR and ARITY. The optional argument
PREFIX is the prefix of the search regexp."
	(let* (;; If prefix is not set then use the default "\\<"
		(prefix 
		    (if (not prefix) "\\<" prefix))
		(regexp (concat prefix functor))
		(i 1)
	) 
	    ;; Build regexp for the search if the arity is > 0
	    (if (= arity 0)
			;; Add that the functor must be at the end of a word. This
			;; does not work if the arity is > 0 since the closing )
			;; is not a word constituent.
			(setq regexp (concat regexp "\\>"))
		    ;; Arity is > 0, add parens and commas
		    (setq regexp (concat regexp "("))
			(while (< i arity)
			    (setq regexp (concat regexp ".+,"))
				(setq i (1+ i)))
			(setq regexp (concat regexp ".+)")))
		;; Search, and return position
		(if (re-search-forward regexp nil t)
			(goto-char (match-beginning 0))
		    (error "Term not found"))
    )
)

(defun mercury-variables-to-anonymous (beg end)
    "Replace all variables within a region BEG to END by anonymous variables."
	(interactive "r")
	(save-excursion
	    (let (
			(oldcase case-fold-search)
		)
		    (setq case-fold-search nil)
			(goto-char end)
			(while (re-search-backward "\\<[A-Z_][a-zA-Z_0-9]*\\>" beg t)
			    (progn
				    (replace-match "_")
					(backward-char)
				))
			(setq case-fold-search oldcase)
		)
	)
)

(defun mercury-set-atom-regexps ()
    "Set the `mercury-atom-char-regexp' and `mercury-atom-regexp' variables.
Must be called after `mercury-build-case-strings'."
	(setq mercury-atom-char-regexp
        (format "[%s%s0-9_$]" mercury-lower-case-string mercury-upper-case-string))
	(setq mercury-atom-regexp
        (format "[%s$]%s*" mercury-lower-case-string mercury-atom-char-regexp))
)

(defun mercury-build-case-strings ()
    "Set `mercury-upper-case-string' and `mercury-lower-case-string'.
Uses the current case-table for extracting the relevant information."
    (let (
		(up_string "")
        (low_string "")
	)
	    ;; Use `map-char-table' if it is defined. Otherwise enumerate all
	    ;; numbers between 0 and 255. `map-char-table' is probably safer.
	    (if (and (not (eq mercury-emacs 'xemacs)) (fboundp 'map-char-table))
			(map-char-table
			    (lambda (key value)
				    (cond 
					    ((and 
						    (eq (int-to-char key) (downcase key))
							(eq (int-to-char key) (upcase key))
						)
						    ;; Do nothing if upper and lower case are the same
						    )
						((eq (int-to-char key) (downcase key))
						    ;; The char is lower case
						    (setq low_string (format "%s%c" low_string key)))
						((eq (int-to-char key) (upcase key))
						    ;; The char is upper case
						    (setq up_string (format "%s%c" up_string key))))
				)
				(current-case-table)
			)
		    ;; `map-char-table' was undefined.
		    (let (
				(key 0)
			)
			    (while (< key 256)
				    (cond 
					    ((and 
						    (eq (int-to-char key) (downcase key))
							(eq (int-to-char key) (upcase key))
						)
						    ;; Do nothing if upper and lower case are the same
						    )
						((eq (int-to-char key) (downcase key))
						    ;; The char is lower case
						    (setq low_string (format "%s%c" low_string key)))
						((eq (int-to-char key) (upcase key))
						    ;; The char is upper case
						    (setq up_string (format "%s%c" up_string key))))
					(setq key (1+ key)))
			))
		;; The strings are single-byte strings
		(setq mercury-upper-case-string up_string)
		(setq mercury-lower-case-string low_string)
    )
)

;; FSF Emacs compatibility: FSF Emacs does not differentiate between
;; ints and chars, or at least these two are interchangeable.
(or 
    (fboundp 'int-to-char)
    ;; Introduced in Emacs 19.29.
    (defun int-to-char (num)
        num
    )
)

;;-------------------------------------------------------------------
;; Menu stuff (both for the editing buffer and for the inferior
;; mercury buffer)
;;-------------------------------------------------------------------

(unless (fboundp 'region-exists-p)
    (defun region-exists-p ()
	    "Non-nil iff the mark is set.  Lobotomized version for Emacsen that do not provide their own."
		(mark)
	)
)

(defun mercury-menu ()
    "Creates the menus for the Mercury editing buffers.
These menus are dynamically created because one may change systems
during the life of an Emacs session, and because FSF Emacs wants them
so by ignoring `easy-menu-add'."
	;; FSF Emacs ignores `easy-menu-add' so the order in which the menus
	;; are defined _is_ important!
	(easy-menu-define 
	    mercury-edit-menu-help 
		(current-local-map)
		"Help menu for the Mercury mode."
		(append
		    (if (eq mercury-emacs 'xemacs) '("Help") '("Mercury-help"))
			'(["Describe mode" describe-mode t])))
	(easy-menu-define 
	    mercury-edit-menu-runtime 
		(current-local-map) 
		"Runtime Mercury commands available from the editing buffer"
		(append 
;; ??????????????????????????????????????????????????????????????????
		    ;; runtime menu name
		    (list "Mercury")
			(list "---"
	        ;;;; TO-DO...
			)))
	(easy-menu-define 
	    mercury-edit-menu-insert-move 
		(current-local-map) 
		"Commands for Mercury code manipulation."
		(append
		    (list "Code"
				["Comment region" comment-region (region-exists-p)]
				["Uncomment region" mercury-uncomment-region (region-exists-p)]
				["Add comment/move to comment" indent-for-comment t])
			(list "---"
				["Insert predicate template" mercury-insert-predicate-template t]
				["Insert next clause head" mercury-insert-next-clause t]
				["Insert predicate spec" mercury-insert-predspec t]
				["Insert module modeline" mercury-insert-module-modeline t]
				"---"
				["Beginning of clause" mercury-beginning-of-clause t]
				["End of clause" mercury-end-of-clause t]
				["Beginning of predicate" mercury-beginning-of-predicate t]
				["End of predicate" mercury-end-of-predicate t]
				"---"
				["Indent line" mercury-indent-line t]
				["Indent region" indent-region (region-exists-p)]
				["Indent predicate" mercury-indent-predicate t]
				["Indent buffer" mercury-indent-buffer t]
				"---"
				["Mark clause" mercury-mark-clause t]
				["Mark predicate" mercury-mark-predicate t]
				["Mark paragraph" mark-paragraph t]
				;"---"
				;["Fontify buffer" font-lock-fontify-buffer t]
            )))

	(easy-menu-add mercury-edit-menu-insert-move)
	(easy-menu-add mercury-edit-menu-runtime)  
	;; Add predicate index menu
	(make-variable-buffer-local 'imenu-create-index-function)
	(setq imenu-create-index-function 'imenu-default-create-index-function)
	;;Milan (this has problems with object methods...)  ###### Does it? (Stefan)
	(setq imenu-prev-index-position-function 'mercury-beginning-of-predicate)
	(setq imenu-extract-index-name-function 'mercury-get-predspec)
  
	(if 
		(and 
		    mercury-imenu-flag
			(< (count-lines (point-min) (point-max)) mercury-imenu-max-lines))
		(imenu-add-to-menubar "Predicates"))
  
	(easy-menu-add mercury-edit-menu-help))

(defun mercury-inferior-menu ()
    "Creates the menus for the Mercury inferior buffer.
This menu is dynamically created because one may change systems during
the life of an Emacs session."
	(easy-menu-define 
	    mercury-inferior-menu-help 
		(current-local-map)
		"Help menu for the Mercury inferior mode."
		(append
		    (if (eq mercury-emacs 'xemacs) '("Help") '("Mercury-help"))
			'(["Describe mode" describe-mode t])))
	(easy-menu-define
	    mercury-inferior-menu-all 
		(current-local-map)
		"Menu for the inferior Mercury buffer."
		(append
		    ;; menu name
		    (list "Mercury")
;;????????????????????????????????????????????????????????????????
			;; runtime
			'("---"
			    ["Interrupt Mercury" comint-interrupt-subjob t]
				["Quit Mercury" comint-quit-subjob t]
				["Kill Mercury" comint-kill-subjob t])))

	(easy-menu-add mercury-inferior-menu-all)
	(easy-menu-add mercury-inferior-menu-help)
)

(add-hook 'mercury-mode-hook 'mercury-menu)
(add-hook 'mercury-inferior-mode-hook 'mercury-inferior-menu)

(add-hook 'mercury-mode-hook '(lambda () (font-lock-mode 1)))
(add-hook 'mercury-inferior-mode-hook '(lambda () (font-lock-mode 1)))


(defun mercury-mode-version ()
    "Echo the current version of Mercury mode in the minibuffer."
	(interactive)
	(message "Using Mercury mode version %s" mercury-mode-version)
)

(provide 'mercury)

;;; mercury.el ends here


;;----------------------------------------------------------------
;; KABUFF


(defun mercury-consult-compile (compilep file &optional first-line)
    "Consult/compile FILE.
If COMPILEP is non-nil, perform compilation, otherwise perform CONSULTING.
COMMAND is a string described by the variabl `mercury-compile-string'.
Optional argument FIRST-LINE is the number of the first line in the compiled
region.

This function must be called from the source code buffer."
    (if mercury-process-flag
		(error "Another Mercury task is running."))
	(let* ( ;-------------------------------------------------------------------
		(buffer (get-buffer-create mercury-compilation-buffer))
		(real-file buffer-file-name)
		(command-string 
		    (mercury-build-mercury-command compilep file real-file first-line))
;;;; Sinnvoll, wenn Mercury ohne Interpreter??
;;  ??????????????????????????????????????????????????????????????????????????
		(process (get-process "mercury"))
		(old-filter (process-filter process))
	) ;-------------------------------------------------------------------------
	    (save-excursion
		    (set-buffer buffer)
			(delete-region (point-min) (point-max))
			(compilation-mode)
			;; Setting up font-locking for this buffer
			(make-local-variable 'font-lock-defaults)
			(setq font-lock-defaults 
				  '(mercury-font-lock-keywords nil nil ((?_ . "w"))))
			(insert command-string "\n")
		)
		(save-selected-window (pop-to-buffer buffer))
		(setq 
		    mercury-process-flag t
			mercury-consult-compile-output ""
			mercury-consult-compile-first-line (if first-line (1- first-line) 0)
			mercury-consult-compile-file file
			mercury-consult-compile-real-file 
			    (if (string= file buffer-file-name) nil real-file))
		(save-excursion
		    (set-buffer buffer)
			(goto-char (point-max))
			(set-process-filter process 'mercury-consult-compile-filter)
			(process-send-string "mercury" command-string)
			;; (mercury-build-mercury-command compilep file real-file first-line))
			(while (and mercury-process-flag (accept-process-output process 10)) ; 10 secs is ok?
			    (sit-for 0.1)
;;;; Sinnvoll, wenn Mercury ohne Interpreter??
;;  ??????????????????????????????????????????????????????????????????????????
				(if (not (get-process "mercury"))
					(setq mercury-process-flag nil)))
			(insert "\nCompilation finished.\n")
			(set-process-filter process old-filter)
		)
	) ;-------------------------------------------------------------------------
)

